# -*- coding: utf-8 -*-
"""
This is a skeleton file that can serve as a starting point for a Python
console script. To run this script uncomment the following lines in the
[options.entry_points] section in setup.cfg:

    console_scripts =
         fibonacci = satproc.skeleton:run

Then run `python setup.py install` which will install the command `fibonacci`
inside your current environment.
Besides console scripts, the header (i.e. until _logger...) of this file can
also be used as template for Python modules.

Note: This skeleton file can be safely removed if not needed!
"""

import argparse
import logging
import sys

from satproc import __version__
from satproc.chips import extract_chips
from satproc.utils import get_raster_band_count

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Extract chips from a raster file, "
        "and optionally generate mask chips",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("raster", nargs="+", help="input raster file(s)")
    parser.add_argument(
        "--size", type=int, default=256, help="size of image tiles, in pixels"
    )
    parser.add_argument(
        "--step-size", type=int, default=None, help="step size (i.e. stride), in pixels (if None, same as size i.e. no overlap)"
    )
    parser.add_argument(
        "--sliding-windows-mode",
        choices=["exact", "whole", "whole_overlap"],
        default="whole_overlap",
        help="mode of sliding windows",
    )
    parser.add_argument("--labels", help="inpul label shapefile")
    parser.add_argument(
        "--label-property",
        default="class",
        help="label property to separate in classes",
    )
    parser.add_argument(
        "--classes", nargs="+", type=str, help="specify classes order in result mask."
    )
    parser.add_argument(
        "--masks",
        "-m",
        nargs="+",
        choices=["extent", "boundary", "distance"],
        default={"extent"},
    )
    parser.add_argument(
        "--mask-type", choices=["single", "class", "instance"], default="class"
    )
    parser.add_argument("--aoi", help="Filter by AOI vector file")
    parser.add_argument(
        "--within",
        dest="within",
        action="store_true",
        help="only create chip is it is within AOI (if provided)",
    )
    parser.add_argument(
        "--no-within",
        dest="within",
        default=False,
        action="store_false",
        help="create chip if it intersects with AOI (if provided)",
    )
    parser.add_argument("-o", "--output-dir", help="output dir", default=".")

    parser.add_argument(
        "--rescale",
        dest="rescale",
        default=False,
        action="store_true",
        help="rescale intensity using percentiles (lower/upper cuts)",
    )
    parser.add_argument(
        "--no-rescale",
        dest="rescale",
        action="store_false",
        help="do not rescale intensity",
    )

    parser.add_argument(
        "--rescale-mode",
        default="percentiles",
        choices=["percentiles", "values", "s2_rgb_extra"],
        help="choose mode of intensity rescaling",
    )

    parser.add_argument(
        "--lower-cut",
        type=float,
        default=2,
        help="(for 'percentiles' mode) lower cut of percentiles "
        "for cumulative count in intensity rescaling",
    )
    parser.add_argument(
        "--upper-cut",
        type=float,
        default=98,
        help="(for 'percentiles' mode) upper cut of percentiles "
        "for cumulative count in intensity rescaling",
    )

    parser.add_argument(
        "--min",
        type=float,
        help="(for 'values' mode) minimum value in intensity rescaling",
    )
    parser.add_argument(
        "--max",
        type=float,
        help="(for 'values' mode) maximum value in intensity rescaling",
    )

    parser.add_argument(
        "-b",
        "--bands",
        nargs="+",
        type=int,
        help="specify band indexes. If type is 'jpg', defaults to (1, 2, 3). "
        "If type is 'tif', defaults to the total band count.",
    )
    parser.add_argument(
        "-t", "--type", help="output chip format", choices=["jpg", "tif"], default="tif"
    )

    parser.add_argument(
        "--write-footprints",
        help="write a GeoJSON file of chip footprints",
        dest="write_footprints",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--no-write-footprints",
        help="do not write a GeoJSON file of chip footprints",
        dest="write_footprints",
        action="store_false",
    )

    parser.add_argument("--crs", help="force CRS of input files")

    parser.add_argument(
        "--skip-existing",
        dest="skip_existing",
        default=True,
        action="store_true",
        help="skip already existing chips (and masks)",
    )
    parser.add_argument(
        "--no-skip-existing",
        dest="skip_existing",
        action="store_false",
        help="do not skip already existing chips (and masks)",
    )

    parser.add_argument(
        "--skip-low-contrast",
        dest="skip_low_contrast",
        default=False,
        action="store_true",
        help="skip image chips with low contrast",
    )
    parser.add_argument(
        "--no-skip-low-contrast",
        dest="skip_low_contrast",
        action="store_false",
        help="do not skip image chips with low contrast",
    )
    parser.add_argument(
        "--skip-with-empty-mask",
        action="store_true",
        default=True,
        help="skip chips with empty mask",
    )
    parser.add_argument(
        "--no-skip-with-empty-mask",
        action="store_false",
        dest="skip_with_empty_mask",
        help="do not skip chips with empty mask",
    )
    parser.add_argument(
        "--extent-no-border",
        action="store_true",
        help="do not include border in extent mask",
    )

    parser.add_argument(
        "--version", action="version", version="satproc {ver}".format(ver=__version__)
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args), parser


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args, parser = parse_args(args)
    setup_logging(args.loglevel)

    bands = args.bands
    if not bands:
        if args.type == "jpg":
            bands = [1, 2, 3]
        else:
            some_raster = args.raster[0]
            bands = list(range(1, get_raster_band_count(some_raster) + 1))

    if type == "jpg" and len(bands) != 3:
        parser.error(
            f"--type is jpg, but --bands does not have 3 band indexes: {bands}"
        )

    rescale_mode = args.rescale_mode if args.rescale else None
    if rescale_mode == "percentiles":
        rescale_range = (args.lower_cut, args.upper_cut)
        _logger.info("Rescale intensity with percentiles %s", rescale_range)
    elif rescale_mode == "values":
        rescale_range = (args.min, args.max)
        _logger.info("Rescale intensity with values %s", rescale_range)
    elif rescale_mode == "s2_rgb_extra":
        rescale_range = (args.lower_cut, args.upper_cut)
        _logger.info(
            "Rescale intensity of first 3 bands (RGB) to (0, 0.3) "
            "and the rest of the bands with percentiles %s",
            rescale_range,
        )
    else:
        rescale_range = None
        _logger.info("No rescale intensity")

    _logger.info("Extract chips")

    extract_chips(
        args.raster,
        aoi=args.aoi,
        labels=args.labels,
        label_property=args.label_property,
        masks=args.masks,
        mask_type=args.mask_type,
        extent_no_border=args.extent_no_border,
        rescale_mode=rescale_mode,
        rescale_range=rescale_range,
        bands=bands,
        chip_type=args.type,
        within=args.within,
        write_footprints=args.write_footprints,
        classes=args.classes,
        crs=args.crs,
        skip_existing=args.skip_existing,
        skip_low_contrast=args.skip_low_contrast,
        skip_with_empty_mask=args.skip_with_empty_mask,
        size=args.size,
        step_size=args.step_size,
        windows_mode=args.sliding_windows_mode,
        output_dir=args.output_dir,
    )


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



import logging
import os

import fiona
import numpy as np
import rasterio
from rasterio.crs import CRS
from shapely.geometry import box, shape
from shapely.ops import unary_union
from skimage import exposure
from skimage.io import imsave
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from satproc.masks import (
    all_masks_empty,
    multiband_chip_mask_by_classes,
    prepare_label_shapes,
    write_window_masks,
)
from satproc.utils import rescale_intensity, sliding_windows, write_chips_geojson

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def extract_chips(
    rasters,
    aoi=None,
    labels=None,
    label_property="class",
    masks={"extent"},
    mask_type="class",
    extent_no_border=False,
    rescale_mode=None,
    rescale_range=None,
    bands=None,
    chip_type="tif",
    write_footprints=False,
    classes=None,
    crs=None,
    skip_existing=True,
    within=False,
    windows_mode="whole_overlap",
    skip_low_contrast=False,
    skip_with_empty_mask=True,
    *,
    size,
    step_size,
    output_dir,
):
    if mask_type not in ("single", "class"):
        raise RuntimeError(f"mask type '{mask_type}' not implemented")

    if aoi:
        _logger.info("Prepare AOI shape")
        aoi_poly = prepare_aoi_shape(aoi)
    else:
        aoi_poly = None

    if labels:
        _logger.info("Prepare label shapes")
        polys_dict = prepare_label_shapes(
            labels, mask_type=mask_type, label_property=label_property, classes=classes
        )
    else:
        polys_dict = None

    if step_size is None:
        step_size = size

    with logging_redirect_tqdm():
        for raster in tqdm(rasters, desc="Rasters", ascii=True):
            extract_chips_from_raster(
                raster,
                size=size,
                step_size=step_size,
                rescale_mode=rescale_mode,
                rescale_range=rescale_range,
                bands=bands,
                output_dir=output_dir,
                chip_type=chip_type,
                within=within,
                write_footprints=write_footprints,
                crs=crs,
                labels=labels,
                label_property=label_property,
                classes=classes,
                masks=masks,
                mask_type=mask_type,
                extent_no_border=extent_no_border,
                aoi_poly=aoi_poly,
                polys_dict=polys_dict,
                windows_mode=windows_mode,
                skip_existing=skip_existing,
                skip_low_contrast=skip_low_contrast,
                skip_with_empty_mask=skip_with_empty_mask,
            )


def extract_chips_from_raster(
    raster,
    rescale_mode=None,
    rescale_range=None,
    bands=None,
    chip_type="tif",
    write_footprints=False,
    labels=None,
    label_property="class",
    mask_type="class",
    masks={"extent"},
    classes=None,
    crs=None,
    skip_existing=True,
    within=False,
    aoi_poly=None,
    polys_dict=None,
    windows_mode="whole_overlap",
    skip_low_contrast=False,
    skip_with_empty_mask=True,
    extent_no_border=False,
    *,
    size,
    step_size,
    output_dir,
):
    if mask_type not in ("single", "class"):
        raise RuntimeError(f"mask type '{mask_type}' not implemented")

    # If mask type is single, there is a single class
    if mask_type == "single":
        classes = None

    if extent_no_border and "extent" not in masks:
        _logger.warn(
            (
                "You specified `no_border` option but will be ignored "
                "(no `extent` mask specified)"
            )
        )

    if skip_existing:
        _logger.info("Will skip existing files")

    basename, _ = os.path.splitext(os.path.basename(raster))

    with rasterio.open(raster) as ds:
        _logger.info("Raster size: %s", (ds.width, ds.height))

        if any(b > ds.count for b in bands):
            raise RuntimeError(
                f"Raster has {ds.count} bands, "
                f"but you asked to use {bands} band indexes"
            )

        if bands is None:
            bands = list(range(1, min(ds.count, 3) + 1))

        _logger.info((
            f"Building windows of size {size}, step size {step_size}"
            f"{' (no overlap)' if step_size >= size else ''}"
        ))
        win_size = (size, size)
        win_step_size = (step_size, step_size)
        windows = list(
            sliding_windows(
                win_size, win_step_size, ds.width, ds.height, mode=windows_mode
            )
        )
        _logger.info("Total windows: %d", len(windows))

        _logger.info("Building window shapes")
        window_shapes = [
            box(*rasterio.windows.bounds(w, ds.transform)) for w, _ in windows
        ]
        window_and_shapes = zip(windows, window_shapes)

        # Filter windows by AOI shape
        if aoi_poly:
            _logger.info("Filtering windows by AOI")
            _logger.info('Using "%s" function', "within" if within else "intersects")

            def filter_fn(w, aoi):
                if within:
                    return w.within(aoi)
                else:
                    return w.intersects(aoi)

            window_and_shapes = [
                (w, s) for w, s in window_and_shapes if filter_fn(s, aoi_poly)
            ]
            _logger.info("Total windows after filtering: %d", len(window_and_shapes))

        meta = ds.meta.copy()
        if crs:
            meta["crs"] = CRS.from_string(crs)
        if rescale_mode:
            # If rescaling, set nodata=0 (will rescale to uint8 1-255)
            meta["nodata"] = 0

        chips = []
        for c, ((window, (i, j)), win_shape) in tqdm(
            list(enumerate(window_and_shapes)),
            desc=f"{os.path.basename(raster)} windows",
            ascii=True,
        ):
            _logger.debug("%s %s", window, (i, j))

            name = f"{basename}_{i}_{j}.{chip_type}"
            img_path = os.path.join(output_dir, "images", name)
            mask_paths = {kind: os.path.join(output_dir, kind, name) for kind in masks}

            # Gather list of required files
            required_files = {img_path}
            if labels:
                required_files = required_files | set(mask_paths.values())

            # If all files already exist and we are skipping existing files, continue
            if skip_existing and all(os.path.exists(p) for p in required_files):
                continue

            img = ds.read(window=window)
            img = np.nan_to_num(img)
            img = np.array([img[b - 1, :, :] for b in bands])

            if rescale_mode:
                img = rescale_intensity(img, rescale_mode, rescale_range)

            if skip_low_contrast and exposure.is_low_contrast(img):
                continue

            # Write mask file
            if labels:
                keys = classes if classes is not None else polys_dict.keys()
                mask_imgs = multiband_chip_mask_by_classes(
                    classes=keys,
                    transform=ds.transform,
                    window=window,
                    polys_dict=polys_dict,
                    extent_mask_path=mask_paths.get("extent"),
                    boundary_mask_path=mask_paths.get("boundary"),
                    distance_mask_path=mask_paths.get("distance"),
                    label_property=label_property,
                    extent_no_border=extent_no_border,
                )

                # If all masks are empty, skip this window
                if skip_with_empty_mask and all_masks_empty(mask_imgs):
                    continue

                write_window_masks(
                    mask_imgs, window=window, metadata=meta, transform=ds.transform
                )

            chip = (win_shape, (c, i, j))
            chips.append(chip)

            # Write image file
            if chip_type == "tif":
                write_tif(
                    img,
                    img_path,
                    window=window,
                    meta=meta.copy(),
                    transform=ds.transform,
                    bands=bands,
                    skip_low_contrast=skip_low_contrast,
                )
            else:
                write_image(
                    img,
                    img_path,
                    skip_low_contrast=skip_low_contrast,
                )

        if write_footprints:
            geojson_path = os.path.join(output_dir, "{}.geojson".format(basename))
            _logger.info("Write chips footprints GeoJSON at %s", geojson_path)
            write_chips_geojson(
                geojson_path,
                chips,
                chip_type=chip_type,
                crs=str(meta["crs"]),
                basename=basename,
            )


def write_image(img, path, *, percentiles=None, skip_low_contrast=False):
    rgb = np.dstack(img[:3, :, :]).astype(np.uint8)
    if skip_low_contrast and exposure.is_low_contrast(rgb):
        return False
    os.makedirs(os.path.dirname(path), exist_ok=True)
    imsave(path, rgb)
    return True


def write_tif(img, path, *, skip_low_contrast=False, window, meta, transform, bands):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    meta.update(
        {
            "driver": "GTiff",
            "dtype": img.dtype,
            "height": window.height,
            "width": window.width,
            "transform": rasterio.windows.transform(window, transform),
            "count": len(bands),
        }
    )
    img = np.array([img[b - 1, :, :] for b in bands])
    with rasterio.open(path, "w", **meta) as dst:
        dst.write(img)
    return True


def get_shape(feature):
    """Get shape geometry from feature

    Parameters
    ----------
    feature : dict
        Feature as read from Fiona

    Returns
    -------
    shapely.geometry.BaseGeometry

    """
    geom = feature["geometry"]
    try:
        return shape(geom)
    except Exception as err:
        _logger.warn("Failed to get shape from feature %s: %s", feature, err)
        return None


def prepare_aoi_shape(aoi):
    with fiona.open(aoi) as src:
        aoi_polys = [get_shape(f) for f in src]
        aoi_polys = [shp for shp in aoi_polys if shp and shp.is_valid]
        aoi_poly = unary_union(aoi_polys)
        return aoi_poly



# -*- coding: utf-8 -*-
import argparse
import logging
import os
import sys
from glob import glob

from satproc import __version__
from satproc.postprocess.polygonize import polygonize

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Polygonize raster images into a single vector file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("input_image", nargs="*", help="input raster")
    parser.add_argument("--input-dir", help="directory containing input rasters")
    parser.add_argument("-o", "--output", help="output in GPKG format")
    parser.add_argument("--temp-dir", help="temporary directory")

    parser.add_argument(
        "--threshold",
        "-t",
        type=float,
        default=None,
        help="apply a threshold before polygonizing",
    )
    parser.add_argument(
        "--value",
        type=float,
        default=None,
        help=(
            "value to use on output after thresholding. "
            "If None, use the original value from src."
        ),
    )

    parser.add_argument(
        "--tile-size",
        type=int,
        default=None,
        help="retile input files with specific tile size",
    )

    parser.add_argument(
        "--version", action="version", version="satproc {ver}".format(ver=__version__)
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args), parser


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel,
        stream=sys.stdout,
        format=logformat,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args, parser = parse_args(args)
    setup_logging(args.loglevel)

    input_files = []
    if args.input_image:
        input_files.extend(args.input_image)
    if args.input_dir:
        files = list(glob(os.path.join(args.input_dir, "*.tif")))
        input_files.extend(files)

    if not input_files:
        raise RuntimeError(
            (
                "No input files found. "
                "You should pass individual input_image paths, or use --input-dir."
            )
        )
    _logger.info("Num. input files: %d", len(input_files))

    polygonize(
        input_files=input_files,
        output=args.output,
        temp_dir=args.temp_dir,
        threshold=args.threshold,
        value=args.value,
        tile_size=args.tile_size,
    )


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



import logging
import os
import tempfile
from functools import partial
from glob import glob

import fiona
from shapely.geometry import mapping, shape
from shapely.ops import unary_union
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from satproc.utils import grouper, map_with_threads, run_command

_logger = logging.getLogger(__name__)


def gdal_polygonize(src, dst):
    run_command(f"gdal_polygonize.py {src} {dst}")


def apply_threshold(src, dst, value=None, *, threshold):
    """
    Output source values (probabilities) instead of simply a binary mask

    Make sure nodata=0, so that gdal_polygonize step ignores pixels under
    threshold.

    Parameters
    ----------
    src : str
        path to input raster
    dst : str
        path to output raster
    threshold : int
        threshold value
    value : int
        value to use on raster output when values are over threshold
        if None, use the original value from src

    """
    # Rescale to 256
    threshold = threshold * 256

    if not value:
        value = "A"
    run_command(
        "gdal_calc.py "
        f'--calc "(A >= {threshold}) * {value}" '
        f"-A {src} "
        "--NoDataValue 0 "
        f"--outfile {dst}"
    )


def process_image(image, value=None, *, tmpdir, threshold):
    src = image
    if threshold:
        src = os.path.join(tmpdir, os.path.basename(image))
        apply_threshold(src=image, dst=src, threshold=threshold, value=value)
    name, _ = os.path.splitext(os.path.basename(image))
    dst = os.path.join(tmpdir, f"{name}.gpkg")
    gdal_polygonize(src, dst)


def merge_vector_files(*, input_dir, output, tmpdir):
    srcs = list(glob(os.path.join(input_dir, "*.gpkg")))
    src_groups = list(enumerate(grouper(srcs, n=1000)))
    groups_dir = os.path.join(tmpdir, "groups")
    os.makedirs(groups_dir, exist_ok=True)

    def merge_chip_vector_files(enumerated_srcs, *, output_dir):
        i, srcs = enumerated_srcs
        srcs = [f for f in srcs if f]
        output = os.path.join(groups_dir, f"{i}.gpkg")
        run_command(
            f"ogrmerge.py -overwrite_ds -single "
            f'-f GPKG -o {output} {" ".join(srcs)}',
            quiet=False,
        )
        return output

    # First, merge groups of vector files using ogrmerge.py in parallel
    output_dir = os.path.join(tmpdir, "temp")
    worker = partial(merge_chip_vector_files, output_dir=output_dir)
    map_with_threads(src_groups, worker, desc="Merge chips into groups")

    # Second, merge ogrmerge results using ogr2ogr into a single file
    group_paths = glob(os.path.join(groups_dir, "*.gpkg"))
    merged_path = os.path.join(output_dir, "merged", os.path.basename(output))
    os.makedirs(os.path.dirname(merged_path) or ".", exist_ok=True)
    with logging_redirect_tqdm():
        for src in tqdm(group_paths, ascii=True, desc="Merge groups"):
            run_command(
                f"ogr2ogr -f GPKG -update -append {merged_path} {src}", quiet=False
            )

    # Third, dissolve the final vector file
    os.makedirs(os.path.dirname(output) or ".", exist_ok=True)
    with fiona.open(merged_path) as src:
        if len(src) == 0:
            _logger.warn("No shapes. Will not write output file")
            return
        with fiona.open(output, "w", **src.meta) as dst:
            all_shapes = list(
                grouper(
                    (
                        shape(feat["geometry"]).buffer(0)
                        for feat in tqdm(
                            src, total=len(src), ascii=True, desc="Reading shapes"
                        )
                    ),
                    n=10000,
                )
            )

            single_shapes = []
            for group_shapes in tqdm(all_shapes, ascii=True, desc="Dissolve"):
                group_shapes = [s for s in group_shapes if s]
                single_shape = unary_union(group_shapes)
                single_shapes.append(single_shape)

            _logger.info("Dissolve final shapes")
            final_shape = unary_union(single_shapes)
            props = {"DN": 255}
            for s in _multipart_to_single_parts(final_shape):
                dst.write({"geometry": mapping(s), "properties": props})
    _logger.info("%s written", output)


def _multipart_to_single_parts(shp):
    if shp.type == "MultiPolygon":
        for s in shp.geoms:
            yield s
    elif shp.type == "Polygon":
        yield shp
    else:
        raise RuntimeError(
            f"shape should be a Polygon or MultiPolygon but was {shp.type}"
        )


def retile_all(input_files, tile_size, temp_dir):
    tiles_dir = os.path.join(temp_dir, "_tiles")
    for raster in input_files:
        retile(raster, output_dir=tiles_dir, tile_size=tile_size)
    tiles_files = list(glob(os.path.join(tiles_dir, "*.tif")))
    _logger.info(
        "Num. tiles sized %dx%d at %s: %d",
        tile_size,
        tile_size,
        tiles_dir,
        len(tiles_files),
    )
    return tiles_files


def retile(raster, output_dir, tile_size):
    os.makedirs(output_dir, exist_ok=True)
    run_command(
        "gdal_retile.py "
        f"-ps {tile_size} {tile_size} "
        f"-targetDir {output_dir} {raster}"
    )


def polygonize(
    threshold=None,
    value=None,
    temp_dir=None,
    input_files=[],
    input_dir=None,
    tile_size=None,
    *,
    output,
):
    if input_dir:
        input_files = list(glob(os.path.join(input_dir, "*.tif")))

    if not input_files:
        raise ValueError("No input files")

    must_remove_temp_dir = False
    if temp_dir:
        # Make sure directory exists
        os.makedirs(temp_dir, exist_ok=True)
    else:
        # Create a temporary directory if user did not provide one
        must_remove_temp_dir = True
        tmpdir = tempfile.TemporaryDirectory()
        temp_dir = tmpdir.name

    # Retile input images before processing (if tile_size is present)
    if tile_size:
        input_files = retile_all(input_files, tile_size, temp_dir)

    # Process all chip images
    worker = partial(process_image, tmpdir=temp_dir, threshold=threshold, value=value)
    map_with_threads(input_files, worker, desc="Polygonize chips")

    # Merge all vector files into a single one
    merge_vector_files(input_dir=temp_dir, output=output, tmpdir=temp_dir)

    if must_remove_temp_dir:
        tmpdir.cleanup()



# -*- coding: utf-8 -*-
import argparse
import logging
import sys

from satproc import __version__
from satproc.postprocess.smooth import smooth_stitch

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Stitch overlapping chips with spline interpolation",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("input_dir", help="results directory")
    parser.add_argument("-o", "--output-dir", help="output directory")
    parser.add_argument("--power", "-p", help="spline interpolation power exponent", type=float, default=1.5)
    parser.add_argument("--temp-dir", help="temporary directory")

    parser.add_argument(
        "--version", action="version", version="satproc {ver}".format(ver=__version__)
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args), parser


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args, _parser = parse_args(args)
    setup_logging(args.loglevel)

    smooth_stitch(input_dir=args.input_dir, output_dir=args.output_dir, power=args.power, temp_dir=args.temp_dir)


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



import json
import logging
import multiprocessing as mp
import os
import subprocess
import tempfile
from itertools import zip_longest
from multiprocessing.pool import ThreadPool

import fiona
import numpy as np
import pyproj
import rasterio
from packaging import version
from pyproj.crs import CRS
from pyproj.enums import WktVersion
from rasterio.windows import Window
from shapely.geometry import mapping
from shapely.ops import transform
from skimage import exposure
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def sliding_windows(size, step_size, width, height, mode="exact"):
    """Slide a window of +size+ by moving it +step_size+ pixels

    Parameters
    ----------
    size : int
        window size, in pixels
    step_size : int
        step or *stride* size when sliding window, in pixels
    width : int
        image width
    height : int
        image height
    mode : str (default: 'exact')
        either one of 'exact', 'whole', 'whole_overlap'.
        - 'exact': clip windows at borders, if needed
        - 'whole': only whole windows
        - 'whole_overlap': only wohle windows, allow overlapping windows at borders.

    Yields
    ------
    Tuple[Window, Tuple[int, int]]
        a pair of Window and a pair of position (i, j)

    """
    w, h = size
    sw, sh = step_size

    whole = mode in ("whole", "whole_overlap")
    end_i = height - h if whole else height
    end_j = width - w if whole else width

    last_pos_i, last_pos_j = 0, 0
    for pos_i, i in enumerate(range(0, end_i, sh)):
        for pos_j, j in enumerate(range(0, end_j, sw)):
            real_w = w if whole else min(w, abs(width - j))
            real_h = h if whole else min(h, abs(height - i))
            yield Window(j, i, real_w, real_h), (pos_i, pos_j)
            last_pos_i, last_pos_j = pos_i, pos_j

    if mode == "whole_overlap" and (height % sh != 0 or width % sw != 0):
        for pos_i, i in enumerate(range(0, height - h, sh)):
            yield Window(width - w, i, w, h), (
                pos_i,
                last_pos_j + 1,
            )
        for pos_j, j in enumerate(range(0, width - w, sw)):
            yield Window(j, height - h, w, h), (
                last_pos_i + 1,
                pos_j,
            )
        yield Window(width - w, height - h, w, h), (last_pos_i + 1, last_pos_j + 1)


def rescale_intensity(image, rescale_mode, rescale_range):
    """
    Calculate percentiles from a range cut and
    rescale intensity of image to byte range

    Parameters
    ----------
    image : numpy.ndarray
        image array
    rescale_mode : str
        rescaling mode, either 'percentiles' or 'values'
    rescale_range : Tuple[number, number]
        input range for rescaling

    Returns
    -------
    numpy.ndarray
        rescaled image
    """

    if rescale_mode == "percentiles":
        in_range = np.percentile(image, rescale_range, axis=(1, 2)).T
    elif rescale_mode == "values":
        min_value, max_value = rescale_range
        if min_value is None:
            min_value = np.min(image)
        if max_value is None:
            max_value = np.max(image)
        in_range = np.array([(min_value, max_value) for _ in range(image.shape[0])])
    elif rescale_mode == "s2_rgb_extra":
        in_range = np.percentile(image, rescale_range, axis=(1, 2)).T
        # Override first 3 ranges for (0, 0.3) (Sentinel-2 L2A TCI range)
        in_range[0] = (0, 0.3)
        in_range[1] = (0, 0.3)
        in_range[2] = (0, 0.3)
    else:
        raise RuntimeError(f"unknown rescale_mode {rescale_mode}")

    return np.array(
        [
            exposure.rescale_intensity(
                image[i, :, :], in_range=tuple(in_range[i]), out_range=(1, 255)
            ).astype(np.uint8)
            for i in range(image.shape[0])
        ]
    )


def write_chips_geojson(output_path, chip_pairs, *, chip_type, crs, basename):
    """Write a GeoJSON containing chips polygons as features

    Parameters
    ----------
    output_path : str
        GeoJSON output path
    chip_pairs : Tuple[Shape, Tuple[int, int, int]]
        a pair with the chip polygon geometry, and a tuple of (feature id, x, y)
    chip_type : str
        chip file type extension (e.g. tif, jpg)
    crs : str
        CRS epsg code of chip polygon geometry
    basename : str
        basename of chip files

    Returns
    -------
    None

    """
    if not chip_pairs:
        _logger.warn("No chips to save")
        return

    _logger.info("Write chips geojson")
    os.makedirs(os.path.dirname(output_path), exist_ok=True)

    with open(output_path, "w") as f:
        d = {"type": "FeatureCollection", "features": []}
        if crs != "epsg:4326":
            code = crs.split(":")[1]
            d["crs"] = {
                "type": "name",
                "properties": {"name": f"urn:ogc:def:crs:EPSG::{code}"},
            }
        for i, (chip, (_fi, xi, yi)) in enumerate(chip_pairs):
            filename = f"{basename}_{xi}_{yi}.{chip_type}"
            feature = {
                "type": "Feature",
                "geometry": mapping(chip),
                "properties": {"id": i, "x": xi, "y": yi, "filename": filename},
            }
            d["features"].append(feature)
        f.write(json.dumps(d))


def get_raster_band_count(path):
    """Get raster band count

    Parameters
    ----------
    path : str
        path of the raster image

    Returns
    -------
    int
        band count

    """
    with rasterio.open(path) as src:
        return src.count


def reproject_shape(shp, from_crs, to_crs, project=None):
    """Reproject a shape from `from_crs` to `to_crs`

    Parameters
    ----------
    shp : Shape
        shape to reproject
    from_crs : str
        CRS epsg code of shape geometry
    to_crs : str
        CRS epsg code of reprojected shape geometry
    project : Optional[str]
        a Transformer instance to use for reprojection

    Returns
    -------
    Shape
        reprojected shape

    """
    if from_crs == to_crs:
        return shp
    if project is None:
        project = pyproj.Transformer.from_crs(
            from_crs, to_crs, always_xy=True
        ).transform
    return transform(project, shp)


def proj_crs_from_fiona_dataset(fio_ds):
    return CRS.from_wkt(fio_ds.crs_wkt)


def fiona_crs_from_proj_crs(proj_crs):
    if version.parse(fiona.__gdal_version__) < version.parse("3.0.0"):
        fio_crs = proj_crs.to_wkt(WktVersion.WKT1_GDAL)
    else:
        # GDAL 3+ can use WKT2
        fio_crs = proj_crs.to_wkt()
    return fio_crs


def build_virtual_raster(image_paths, output_path, separate=None, band=None):
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    # For some reason, relative paths wont work, so we get the absolute path of
    # each input image path.
    image_paths = [os.path.abspath(p) for p in image_paths]
    with tempfile.NamedTemporaryFile() as f:
        # Write a list of image files to a temporary file
        for image_path in image_paths:
            f.write(f"{image_path}\n".encode())
        f.flush()
        output_path = os.path.abspath(output_path)
        run_command(
            f"gdalbuildvrt -q -overwrite "
            f"-input_file_list {f.name} "
            f"{'-separate ' if separate else ''}"
            f"{f'-b {band} ' if band else ''}"
            f"{output_path}",
            cwd=os.path.dirname(output_path),
        )


def run_command(cmd, quiet=True, *, cwd=None):
    """Run a shell command

    Parameters
    ----------
    cmd : str
        command to run
    quiet : bool (default: True)
        silent output (stdout and sterr)

    Returns
    -------
    None

    """
    stderr = subprocess.DEVNULL if quiet else None
    stdout = subprocess.DEVNULL if quiet else None
    subprocess.run(cmd, shell=True, stderr=stderr, stdout=stdout, check=True, cwd=cwd)


def map_with_threads(items, worker, num_jobs=None, total=None, desc=None):
    """Map a worker function to an iterable of items, using a thread pool

    Parameters
    ----------
    items : iterable
        items to map
    worker : Function
        worker function to apply to each item
    num_jobs : int
        number of threads to use
    total : int (optional)
        total number of items (for the progress bar)
    desc : str (optional)
        description of the task (for the progress bar)

    Returns
    -------
    None

    """
    if not total:
        total = len(items)
    if not num_jobs:
        num_jobs = mp.cpu_count()
    with ThreadPool(num_jobs) as pool:
        with logging_redirect_tqdm():
            with tqdm(total=len(items), ascii=True, desc=desc) as pbar:
                for _ in enumerate(pool.imap_unordered(worker, items)):
                    pbar.update()



# -*- coding: utf-8 -*-
import argparse
import logging
import os
import sys
from glob import glob

from satproc import __version__
from satproc.postprocess.generalize import generalize

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Generalize vector files by simplifying and/or smoothing polygons",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("input_vector", nargs="*", help="input vector file")
    parser.add_argument("--input-dir", help="directory containing input rasters")
    parser.add_argument("-o", "--output-dir", help="output directory")

    parser.add_argument(
        "-tcrs",
        "--target-crs",
        default=None,
        help="reproject to another CRS before processing (EPSG code)",
    )

    parser.add_argument(
        "--simplify",
        choices=["douglas"],
        default="douglas",
        help="simplification method",
    )
    parser.add_argument(
        "--douglas-tolerance",
        type=float,
        default=0.1,
        help="Douglas-Peucker tolerance",
    )

    parser.add_argument(
        "--smooth",
        choices=["chaikin"],
        default=None,
        help="smoothing method",
    )
    parser.add_argument(
        "--chaikin-refinements",
        default=5,
        type=int,
        help="number of Chaikin refinements",
    )

    parser.add_argument(
        "--version", action="version", version="satproc {ver}".format(ver=__version__)
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args), parser


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel,
        stream=sys.stdout,
        format=logformat,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args, parser = parse_args(args)
    setup_logging(args.loglevel)

    input_files = []
    if args.input_vector:
        input_files.extend(args.input_vector)
    if args.input_dir:
        files = list(glob(os.path.join(args.input_dir, "*")))
        input_files.extend(files)

    if not input_files:
        raise RuntimeError(
            (
                "No input files found. "
                "You should pass individual input_vector paths, or use --input-dir."
            )
        )
    _logger.info("Num. input files: %d", len(input_files))

    generalize(
        input_files=input_files,
        output_dir=args.output_dir,
        target_crs=args.target_crs,
        simplify=args.simplify,
        douglas_tolerance=args.douglas_tolerance,
        smooth=args.smooth,
        chaikins_refinements=args.chaikin_refinements,
    )


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
"""
This is a skeleton file that can serve as a starting point for a Python
console script. To run this script uncomment the following lines in the
[options.entry_points] section in setup.cfg:

    console_scripts =
         fibonacci = satproc.skeleton:run

Then run `python setup.py install` which will install the command `fibonacci`
inside your current environment.
Besides console scripts, the header (i.e. until _logger...) of this file can
also be used as template for Python modules.

Note: This skeleton file can be safely removed if not needed!
"""

import argparse
import logging
import sys

from satproc import __version__
from satproc.masks import make_masks

__author__ = "Damián Silvani"
__copyright__ = "Damián Silvani"
__license__ = "mit"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Build masks from chip raster files",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("raster", nargs="+", help="input raster file(s)")

    parser.add_argument(
        "--masks",
        "-m",
        nargs="+",
        choices=["extent", "boundary", "distance"],
        default={"extent"},
    )
    parser.add_argument(
        "--mask-type", choices=["single", "class", "instance"], default="class"
    )
    parser.add_argument("--labels", required=True, help="inpul label shapefile")
    parser.add_argument(
        "--label-property",
        default="class",
        help="label property to separate in classes",
    )
    parser.add_argument(
        "--classes", nargs="+", type=str, help="specify classes order in result mask."
    )
    parser.add_argument(
        "--extent-no-border",
        action="store_true",
        help="do not include border in extent mask",
    )
    parser.add_argument("-o", "--output-dir", help="output dir", default=".")

    parser.add_argument(
        "--version", action="version", version="satproc {ver}".format(ver=__version__)
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args), parser


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args, _parser = parse_args(args)
    setup_logging(args.loglevel)

    make_masks(
        args.raster,
        output_dir=args.output_dir,
        masks=args.masks,
        mask_type=args.mask_type,
        labels=args.labels,
        label_property=args.label_property,
        classes=args.classes,
        extent_no_border=args.extent_no_border,
    )


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



import logging

import numpy as np
import rasterio
from skimage import exposure
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from satproc.utils import sliding_windows

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)

# TODO add win size and step size


def read_window(ds, window):
    """Read from a rasterio dataset using a window

    NaNs are coerced to zero.

    Parameters
    ----------
    ds : rasterio.Dataset
        input dataset
    window : rasterio.windows.Window
        window to read from

    Returns
    -------
    numpy.ndarray
        image data on window

    """

    img = ds.read(window=window)
    img = np.nan_to_num(img)
    return np.dstack(img)


def write_window(img, ds, window):
    """Write array to raster on window

    Parameters
    ----------
    img : numpy.ndarray
        image array
    ds : rasterio.Dataset
        dataset to write to (must be opened with write access)
    window : rasterio.windows.Window
        window to write to

    Returns
    -------
    None

    """
    new_img = np.array([img[:, :, i] for i in range(img.shape[2])])
    ds.write(new_img, window=window)


def match_histograms(src_path, dst_path, size=128, step_size=128, *, reference_path):
    """Match histograms of an image using another one as reference

    Parameters
    ----------
    src_path : str
        path to input raster
    dst_path : str
        path to output raster
    size : int
        size of windows
    step_size : int
        step size, when sliding windows
    reference_path : str
        path to the reference raster

    Returns
    -------
    None

    """
    with rasterio.open(src_path) as src:
        profile = src.profile.copy()
        windows = list(
            sliding_windows(
                (size, size),
                (step_size, step_size),
                src.width,
                src.height,
            )
        )

        with rasterio.open(reference_path) as ref:
            with rasterio.open(dst_path, "w", **profile) as dst:
                with logging_redirect_tqdm():
                    for c, (win, (i, j)) in tqdm(
                        list(enumerate(windows)), ascii=True, desc="Match histograms"
                    ):
                        _logger.debug("%s %s", win, (i, j))

                        img = read_window(src, win)
                        ref_img = read_window(ref, win)

                        matched_img = exposure.match_histograms(img, ref_img)
                        write_window(matched_img, dst, win)



import logging
import os
from collections import defaultdict

import cv2
import fiona
import numpy as np
import rasterio
from rasterio.features import rasterize
from shapely.geometry import shape
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def make_masks(
    rasters,
    *,
    output_dir,
    labels,
    label_property="class",
    classes=None,
    mask_type="class",
    masks={"extent"},
    extent_no_border=False,
):
    if mask_type not in ("single", "class"):
        raise RuntimeError(f"mask type '{mask_type}' not implemented")

    # If mask type is single, there is a single class
    if mask_type == "single":
        classes = None

    if extent_no_border and "extent" not in masks:
        _logger.warn(
            (
                "You specified `no_border` option but will be ignored "
                "(no `extent` mask specified)"
            )
        )

    polys_dict = classify_polygons(labels, label_property, classes)

    with logging_redirect_tqdm():
        for raster in tqdm(rasters, ascii=True, desc="Rasters"):
            name, _ = os.path.splitext(os.path.basename(raster))
            with rasterio.open(raster) as src:
                profile = src.profile.copy()
                transform = src.transform
                b = src.bounds

                # Calculate window from image bounds
                window = rasterio.windows.from_bounds(
                    b.left,
                    b.bottom,
                    b.right,
                    b.top,
                    src.transform,
                    src.height,
                    src.width,
                )

            paths = {
                kind: os.path.join(output_dir, kind, f"{name}.tif") for kind in masks
            }
            keys = classes if classes is not None else polys_dict.keys()

            masks = multiband_chip_mask_by_classes(
                classes=keys,
                transform=transform,
                window=window,
                polys_dict=polys_dict,
                label_property=label_property,
                extent_no_border=extent_no_border,
                extent_mask_path=paths.get("extent"),
                boundary_mask_path=paths.get("boundary"),
                distance_mask_path=paths.get("distance"),
            )

            write_window_masks(
                masks, window=window, metadata=profile, transform=transform
            )


def multiband_chip_mask_by_classes(
    classes,
    transform,
    window,
    label_property,
    polys_dict=None,
    label_path=None,
    extent_mask_path=None,
    boundary_mask_path=None,
    distance_mask_path=None,
    extent_no_border=False,
):
    mb_extent, mb_boundary, mb_distance = [], [], []

    if polys_dict is None and label_path is not None:
        polys_dict = classify_polygons(label_path, label_property, classes)

    for k in classes:
        polys = polys_dict.get(k, [])
        extent_mask, bound_mask, dist_mask = mask_from_polygons(
            polys,
            win=window,
            t=transform,
            extent_no_border=extent_no_border,
            boundary_mask=boundary_mask_path,
            distance_mask=distance_mask_path,
        )
        mb_extent.append(extent_mask)
        mb_boundary.append(bound_mask)
        mb_distance.append(dist_mask)

    masks = zip(
        [mb_extent, mb_boundary, mb_distance],
        [extent_mask_path, boundary_mask_path, distance_mask_path],
    )
    masks = [(path, mask) for mask, path in masks if path is not None]

    return masks


def all_masks_empty(masks):
    """Check if all masks are empty"""
    for _, mask in masks:
        for arr in mask:
            if np.any(arr):
                return False
    return True


def write_window_masks(masks, *, window, metadata, transform):
    """Write window masks to files"""
    for path, mask in masks:
        kwargs = metadata.copy()
        kwargs.update(
            driver="GTiff",
            dtype=rasterio.uint8,
            count=len(mask),
            nodata=0,
            transform=rasterio.windows.transform(window, transform),
            width=window.width,
            height=window.height,
        )

        os.makedirs(os.path.dirname(path), exist_ok=True)
        with rasterio.open(path, "w", **kwargs) as dst:
            for i in range(len(mask)):
                dst.write(mask[i], i + 1)


def mask_from_polygons(
    polygons,
    *,
    win,
    t,
    extent_no_border=True,
    boundary_mask=None,
    distance_mask=None,
):
    """Generate a binary mask array from a set of polygon

    It can also generate a distance transform mask

    Parameters
    ----------
    polygons : List[Union[Polygon, MultiPolygon]]
        list of polygon or multipolygon geometries
    win : rasterio.windows.Window
        window
    t : rasterio.transform.Affine
        affine transform
    extent_no_border : bool
        if True, the extent mask will not include the border of the polygon
    boundary_mask : bool
        whether to generate boundary (edges) mask
    distance_mask : bool
        whether to generate a distance mask

    Returns
    -------
    numpy.ndarray

    """
    transform = rasterio.windows.transform(win, t)
    win_shape = (int(win.height), int(win.width))
    bound_mask, dist_mask = None, None

    if polygons is None or len(polygons) == 0:
        mask = np.zeros(win_shape, dtype=np.uint8)
        if boundary_mask:
            bound_mask = mask.copy()
        if distance_mask:
            dist_mask = mask.copy()
        return mask, bound_mask, dist_mask

    mask = rasterize(
        polygons,
        win_shape,
        default_value=255,
        transform=transform,
    )
    if boundary_mask or distance_mask or extent_no_border:
        lines = _get_linestrings_from_polygons(polygons)
        bound_mask = rasterize(lines, win_shape, default_value=255, transform=transform)
        if extent_no_border or distance_mask:
            mask_no_bounds = mask.copy()
            mask_no_bounds[bound_mask != 0] = 0
            if extent_no_border:
                mask = mask_no_bounds
            if distance_mask:
                dist_mask = cv2.distanceTransform(
                    mask_no_bounds, cv2.DIST_L2, 3
                ).astype(np.uint8)
    return mask, bound_mask, dist_mask


def _get_linestrings_from_polygons(polys):
    for poly in polys:
        boundary = poly.boundary
        if boundary.type == "MultiLineString":
            for line in boundary.geoms:
                yield line
        else:
            yield boundary


def prepare_label_shapes(
    labels, mask_type="class", label_property="class", classes=None
):
    if mask_type in ("class", "single"):
        polys_dict = classify_polygons(labels, label_property, classes)
        return polys_dict
    else:
        raise RuntimeError(f"mask type '{mask_type}' not supported")


def classify_polygons(labels, label_property, classes):
    with fiona.open(labels) as blocks:
        _logger.info("Found %d labels on label file %s", len(blocks), labels)
        if classes is not None and label_property:
            if label_property not in blocks.schema["properties"]:
                raise ValueError(
                    f"label property '{label_property}' not present in labels schema"
                )
        class_counts = defaultdict(int)
        missing_classes = set()
        polys = defaultdict(list)
        for block in blocks:
            if classes is None:
                c = "_any"
            else:
                c = str(block["properties"][label_property])
                class_counts[c] += 1
                if c not in classes:
                    missing_classes.add(c)
                    continue
            try:
                geom = shape(block["geometry"])
            except RuntimeError:
                _logger.warning(
                    "Failed to get geometry shape for feature: %s", block
                )
                continue
            polys[c].append(geom)
    if missing_classes:
        _logger.warn(
            f"Some features with classes {missing_classes} were ignored as they are not in the specified classes list"
        )
        for c in missing_classes:
            _logger.warn(f"Num. features of class '{c}': {class_counts[c]}")
    return polys



import logging
import os
from functools import partial
from glob import glob

import numpy as np
import rasterio

from satproc.utils import map_with_threads

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def get_max_prob(p):
    with rasterio.open(p) as src:
        img = src.read()
        return np.max(img)


def filter_chip(src, *, threshold, output_dir):
    if get_max_prob(src) >= threshold:
        dst = os.path.join(output_dir, os.path.basename(src))
        os.makedirs(output_dir, exist_ok=True)
        if not os.path.exists(dst):
            os.symlink(src, dst)


def filter_by_max_prob(input_dir, output_dir, threshold):
    threshold = round(threshold * 255)
    files = glob(os.path.join(input_dir, "*"))
    worker = partial(filter_chip, output_dir=output_dir, threshold=threshold)
    map_with_threads(files, worker)



import logging

import numpy as np
import rasterio
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from satproc.utils import sliding_windows

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def get_min_max(img, window_size=512):
    """Return minimum and maximum values on array, in blocks

    Parameters
    ----------
    img : numpy.ndarray
        image array
    window_size : int
        size of window (default: 512)

    Returns
    -------
    Tuple[float, float]
        minimum and maximum values

    """
    mins, maxs = [], []
    with rasterio.open(img) as src:
        win_size = (window_size, window_size)
        windows = list(sliding_windows(win_size, win_size, src.width, src.height))
        with logging_redirect_tqdm():
            for _, (window, (i, j)) in tqdm(
                list(enumerate(windows)), desc="Get min-max", ascii=True
            ):
                img = src.read(window=window)
                mins.append(
                    np.array([np.nanmin(img[i, :, :]) for i in range(src.count)])
                )
                maxs.append(
                    np.array([np.nanmax(img[i, :, :]) for i in range(src.count)])
                )
    min_value = np.nanmin(np.array(mins), axis=0)
    max_value = np.nanmax(np.array(maxs), axis=0)
    return min_value, max_value


def minmax_scale(img, *, min_values, max_values):
    """
    Scale bands of image separately, to range 0..1

    Parameters
    ----------
    img : numpy.ndarray
        image array
    min_values : List[float]
        minimum values for each band
    max_values : List[float]
        maximum values for each band

    Returns
    -------
    numpy.ndarray
        rescaled image

    """
    n_bands = img.shape[0]
    return np.array(
        [
            (img[i, :, :] - min_values[i]) / (max_values[i] - min_values[i])
            for i in range(n_bands)
        ]
    )


def scale(input_img, output_img, window_size=512):
    """
    Read a raster, rescale each band with min-max values, and save as another raster

    Parameters
    ----------
    input_img : str
        path to input image
    output_img : str
        path to output image
    window_size : int
        size of window

    Returns
    -------
    None

    """
    min_values, max_values = get_min_max(input_img, window_size=window_size)

    with rasterio.open(input_img) as src:
        win_size = (window_size, window_size)
        windows = list(sliding_windows(win_size, win_size, src.width, src.height))
        profile = src.profile.copy()
        profile.update(dtype=np.float32)
        with rasterio.open(output_img, "w", **profile) as dst:
            with logging_redirect_tqdm():
                for _, (window, (i, j)) in tqdm(
                    list(enumerate(windows)), ascii=True, desc="Scale"
                ):
                    _logger.debug("%s %s", window, (i, j))
                    img = src.read(window=window)
                    new_img = minmax_scale(
                        img, min_values=min_values, max_values=max_values
                    )
                    dst.write(new_img, window=window)



import logging
import os

import fiona
import numpy as np
from pyproj.crs import CRS
from shapely.geometry import Polygon, mapping, shape
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from ..utils import (
    fiona_crs_from_proj_crs,
    proj_crs_from_fiona_dataset,
    reproject_shape,
)

_logger = logging.getLogger(__name__)


def generalize(
    *,
    input_files,
    output_dir,
    target_crs=None,
    simplify="douglas",
    douglas_tolerance=0.1,
    smooth=None,
    chaikins_refinements=5,
):
    if target_crs:
        target_crs = CRS.from_user_input(target_crs)
        _logger.info("Going to reproject shapes to %s", target_crs.to_string())
        if target_crs.is_geographic:
            _logger.warn(
                (
                    "Target CRS is geographic (%s), so keep in mind "
                    "that distance-related parameters are in degrees. "
                    "Consider reprojecting to a projected local "
                    "coordinate system instead for accurate results."
                ),
                target_crs.to_string(),
            )

    with logging_redirect_tqdm():
        for input_file in tqdm(input_files, ascii=True, desc="Generalize"):
            output_file = os.path.join(output_dir, os.path.basename(input_file))
            os.makedirs(os.path.dirname(output_file), exist_ok=True)
            with fiona.open(input_file) as src:
                proj_crs = proj_crs_from_fiona_dataset(src)
                if not target_crs and proj_crs.is_geographic:
                    _logger.warn(
                        (
                            "File %s has a geographic CRS (%s), so keep in mind "
                            "that distance-related parameters are in degrees. "
                            "Consider reprojecting to a projected local "
                            "coordinate system for accurate results."
                        ),
                        input_file,
                        proj_crs.to_string(),
                    )

                dst_meta = src.meta.copy()
                if target_crs:
                    del dst_meta["crs"]
                    dst_meta["crs_wkt"] = fiona_crs_from_proj_crs(target_crs)

                with fiona.open(output_file, "w", **dst_meta) as dst:
                    for feat in tqdm(
                        src, ascii=True, desc=os.path.basename(input_file)
                    ):
                        shp = shape(feat["geometry"])
                        if target_crs:
                            shp = reproject_shape(shp, proj_crs, target_crs)
                        if simplify == "douglas":
                            shp = shp.simplify(
                                tolerance=douglas_tolerance, preserve_topology=True
                            )
                        if smooth == "chaikin":
                            shp = smooth_chaikin(shp, refinements=chaikins_refinements)
                        feat["geometry"] = mapping(shp)
                        dst.write(feat)


# Based on https://stackoverflow.com/a/47255374/1650058
def smooth_chaikin(shp, refinements=5):
    coords = np.array(shp.exterior.coords)

    for _ in range(refinements):
        L = coords.repeat(2, axis=0)
        R = np.empty_like(L)
        R[0] = L[0]
        R[2::2] = L[1:-1:2]
        R[1:-1:2] = L[2::2]
        R[-1] = L[-1]
        coords = L * 0.75 + R * 0.25

    return Polygon(coords)



import logging
import tempfile
import os

import rasterio
import cv2
import numpy as np
from tqdm import tqdm

from satproc.utils import build_virtual_raster

_logger = logging.getLogger(__name__)

MODES = ["gaussian", "median"]


def spatial_filter(*, input_paths, output_path, mode="gaussian", size=5, merge=True):
    if merge:
        temp_dir = tempfile.TemporaryDirectory()
        vrt_path = os.path.join(temp_dir.name, "merged.vrt")
        _logger.info(f"Merge input images into a virtual raster {vrt_path}")
        build_virtual_raster(input_paths, vrt_path)
        input_paths = [vrt_path]

    for input_path in tqdm(input_paths, desc="Apply filter", ascii=True):
        with rasterio.open(input_path) as src:
            if size > src.width or size > src.height:
                raise ValueError(f"Kernel size {size} is larger than image size ({src.width}x{src.height})")

            profile = src.profile.copy()
            profile.update(driver="GTiff")

            with rasterio.open(output_path, "w", **profile) as dst:
                for b in range(1, src.count + 1):
                    img = src.read(b)
                    if mode == "gaussian":
                        new_img = cv2.GaussianBlur(img, (size, size), 0)
                    elif mode == "median":
                        new_img = cv2.medianBlur(img, size)
                    else:
                        raise ValueError(f"Unknown mode: {mode}")
                    dst.write(new_img, b)



import logging
import math
import os
import tempfile
from glob import glob

import numpy as np
import rasterio
import rasterio.mask
import rasterio.merge
import rasterio.windows
import scipy.signal
from rasterio.transform import Affine
from rtree import index
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

logger = logging.getLogger(__name__)


def spline_window(window_size, power=2):
    """
    Squared spline window function:
    https://www.wolframalpha.com/input/?i=y%3Dx**2,+y%3D-(x-2)**2+%2B2,+y%3D(x-4)**2,+from+y+%3D+0+to+2
    """
    intersection = int(window_size / 4)
    wind_outer = (abs(2 * (scipy.signal.triang(window_size))) ** power) / 2
    wind_outer[intersection:-intersection] = 0

    wind_inner = 1 - (abs(2 * (scipy.signal.triang(window_size) - 1)) ** power) / 2
    wind_inner[:intersection] = 0
    wind_inner[-intersection:] = 0

    wind = wind_inner + wind_outer
    wind = wind / np.average(wind)
    return wind


def window_2D(size, power=2, n_channels=1):
    wind = spline_window(size, power)
    wind = np.expand_dims(wind, 0)
    wind = (wind * wind.transpose()) / 4
    wind = np.expand_dims(wind, axis=0)
    return np.repeat(wind, n_channels, axis=0).reshape(n_channels, size, size)


def generate_spline_window_chips(*, image_paths, output_dir, power=2):
    """Interpolates all images using a squared spline window"""
    if not image_paths:
        return []

    logger.info((
        f"Interpolate all images using a squared spline window (power={power}) "
        f"and store chips on {output_dir}"
    ))

    first_image = image_paths[0]
    with rasterio.open(first_image) as src:
        chip_size = src.width
        n_channels = src.count
        assert src.width == src.height

    win = window_2D(size=chip_size, power=power, n_channels=n_channels)
    norm_win = (win - win.min()) / (win.max() - win.min())

    res = []
    with logging_redirect_tqdm():
        for img_path in tqdm(image_paths, ascii=True, desc="Smooth chips"):
            with rasterio.open(img_path) as src:
                profile = src.profile.copy()
                profile.update(dtype=np.float64)
                img = src.read()

            img = img * norm_win

            out_path = os.path.join(output_dir, os.path.basename(img_path))
            os.makedirs(output_dir, exist_ok=True)
            res.append(out_path)
            with rasterio.open(out_path, "w", **profile) as dst:
                for i in range(img.shape[0]):
                    dst.write(img[i, :, :], i + 1)

    return res


def build_bounds_index(image_files):
    """Returns bounds of merged images and builds an R-Tree index"""
    idx = index.Index()
    xs = []
    ys = []
    with logging_redirect_tqdm():
        for i, img_path in tqdm(
            list(enumerate(image_files)), ascii=True, desc="Build bounds R-Tree index"
        ):
            with rasterio.open(img_path) as src:
                left, bottom, right, top = src.bounds
            xs.extend([left, right])
            ys.extend([bottom, top])
            idx.insert(i, (left, bottom, right, top))
    dst_w, dst_s, dst_e, dst_n = min(xs), min(ys), max(xs), max(ys)
    return idx, (dst_w, dst_s, dst_e, dst_n)


def sliding_windows(size, whole=False, step_size=None, *, width, height):
    """Slide a window of +size+ by moving it +step_size+ pixels"""
    if not step_size:
        step_size = size
    w, h = (size, size)
    sw, sh = (step_size, step_size)
    end_i = height - h if whole else height
    end_j = width - w if whole else width
    for pos_i, i in enumerate(range(0, end_i, sh)):
        for pos_j, j in enumerate(range(0, end_j, sw)):
            real_w = w if whole else min(w, abs(width - j))
            real_h = h if whole else min(h, abs(height - i))
            yield rasterio.windows.Window(j, i, real_w, real_h), (pos_i, pos_j)


def merge_chips(images_files, *, win_bounds):
    """Merge by taking mean between overlapping images"""
    datasets = [rasterio.open(p) for p in images_files]
    img, _ = rasterio.merge.merge(datasets, bounds=win_bounds, method="max")
    for ds in datasets:
        ds.close()
    return img


def smooth_stitch(*, input_dir, output_dir, power=1.5, temp_dir=None):
    """
    Takes input directory of overlapping chips, and generates a new directory
    of non-overlapping chips with smooth edges.
    """
    image_paths = glob(os.path.join(input_dir, "*.tif"))
    if not image_paths:
        raise RuntimeError("%s does not contain any .tif file" % (input_dir))

    # Get the profile and affine of some image as template for output image
    first_image = image_paths[0]
    with rasterio.open(first_image) as src:
        profile = src.profile.copy()
        src_res = src.res
        chip_size = src.width
        assert src.width == src.height

    if temp_dir:
        tmpdir = temp_dir
    else:
        tmp_dir = tempfile.TemporaryDirectory()
        tmpdir = tmp_dir.name

    tmp_image_paths = generate_spline_window_chips(
        image_paths=image_paths, output_dir=tmpdir, power=power
    )

    # Get bounds from all images and build R-Tree index
    idx, (dst_w, dst_s, dst_e, dst_n) = build_bounds_index(tmp_image_paths)

    # Get affine transform for complete bounds
    logger.info("Output bounds: %r", (dst_w, dst_s, dst_e, dst_n))
    output_transform = Affine.translation(dst_w, dst_n)
    logger.info("Output transform, before scaling: %r", output_transform)

    output_transform *= Affine.scale(src_res[0], -src_res[1])
    logger.info("Output transform, after scaling: %r", output_transform)

    # Compute output array shape. We guarantee it will cover the output
    # bounds completely. We need this to build windows list later.
    output_width = int(math.ceil((dst_e - dst_w) / src_res[0]))
    output_height = int(math.ceil((dst_n - dst_s) / src_res[1]))

    # Set width and height for output chips, and other attributes
    profile.update(width=chip_size, height=chip_size, tiled=True)

    windows = list(
        sliding_windows(chip_size, width=output_width, height=output_height)
    )
    logger.info("Num. windows: %d", len(windows))

    with logging_redirect_tqdm():
        for win, (i, j) in tqdm(windows, ascii=True, desc="Merge chips"):
            # Get window affine transform and bounds
            win_transform = rasterio.windows.transform(win, output_transform)
            win_bounds = rasterio.windows.bounds(win, output_transform)

            # Get chips that intersect with window
            intersect_chip_paths = [
                tmp_image_paths[i] for i in idx.intersection(win_bounds)
            ]

            if intersect_chip_paths:
                # Merge them with median method
                img = merge_chips(intersect_chip_paths, win_bounds=win_bounds)

                # Write output chip
                profile.update(transform=win_transform)
                output_path = os.path.join(output_dir, f"{i}_{j}.tif")

                os.makedirs(output_dir, exist_ok=True)
                with rasterio.open(output_path, "w", **profile) as dst:
                    for i in range(img.shape[0]):
                        dst.write(img[i, :, :], i + 1)

    if not temp_dir:
        tmp_dir.cleanup()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['satproc', 'satproc.console', 'satproc.postprocess']

package_data = \
{'': ['*']}

install_requires = \
['Fiona>=1.8.21,<2.0.0',
 'Rtree>=0.9.7,<0.10.0',
 'Shapely>=1.8.1,<2.0.0',
 'numpy>=1.22.2,<2.0.0',
 'opencv-python>=4.5.5,<5.0.0',
 'packaging>=21.3,<22.0',
 'pyproj>=3.3.0,<4.0.0',
 'rasterio==1.3b1',
 'scikit-image>=0.19.2,<0.20.0',
 'scipy>=1.8.0,<2.0.0',
 'tqdm>=4.62.3,<5.0.0']

entry_points = \
{'console_scripts': ['satproc_extract_chips = '
                     'satproc.console.extract_chips:run',
                     'satproc_filter = satproc.console.filter:run',
                     'satproc_generalize = satproc.console.generalize:run',
                     'satproc_make_masks = satproc.console.make_masks:run',
                     'satproc_match_histograms = '
                     'satproc.console.match_histograms:run',
                     'satproc_polygonize = satproc.console.polygonize:run',
                     'satproc_scale = satproc.console.scale:run',
                     'satproc_smooth_stitch = '
                     'satproc.console.smooth_stitch:run',
                     'satproc_spatial_filter = '
                     'satproc.console.spatial_filter:run']}

setup_kwargs = {
    'name': 'pysatproc',
    'version': '0.2.0a0',
    'description': 'Python library and CLI tools for processing geospatial imagery for ML',
    'long_description': None,
    'author': 'Damián Silvani',
    'author_email': 'munshkr@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
Apply histogram matching between two rasters.

It manipulates the pixels of an input image so that its histogram matches the
histogram of the reference image. The matching is done independently for each
band, as long as the number of bands is equal in the input image and the
reference.
"""

import argparse
import logging
import sys

from satproc import __version__
from satproc.histogram import match_histograms

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Match histograms between images",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--version",
        action="version",
        version="satproc {ver}".format(ver=__version__),
    )

    parser.add_argument("src", help="input raster")
    parser.add_argument("dst", help="output raster")
    parser.add_argument("-r", "--reference", required=True, help="reference raster")

    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args = parse_args(args)
    setup_logging(args.loglevel)

    match_histograms(
        args.src,
        args.dst,
        reference_path=args.reference,
    )


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
import argparse
import logging
import sys
import os
from glob import glob

from satproc import __version__
from satproc.postprocess.spatial_filter import MODES, spatial_filter

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def check_kernel_size(value):
    ivalue = int(value)
    if ivalue % 2 == 0:
        raise argparse.ArgumentTypeError("must be an odd number")
    return ivalue


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Apply a spatial filter (median, gaussian) to an image",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--version",
        action="version",
        version="satproc {ver}".format(ver=__version__),
    )

    parser.add_argument("input", nargs="*", help="input raster")
    parser.add_argument("--input-dir", help="directory containing input rasters")
    parser.add_argument("--output", "-o", help="output raster")
    parser.add_argument("--mode", "-m", choices=MODES, default="gaussian", help="Filter mode")
    parser.add_argument("--size", "-s", type=check_kernel_size, default=5, help="Size of the filter")
    parser.add_argument(
        "--merge",
        dest="merge",
        default=True,
        action="store_true",
        help="merge all input files into a single virtual raster before filtering",
    )
    parser.add_argument(
        "--no-merge",
        dest="merge",
        default=False,
        action="store_false",
        help="do not merge all input files into a single virtual raster",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )

    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args = parse_args(args)
    setup_logging(args.loglevel)

    input_paths = []
    if args.input:
        input_paths.extend(args.input)
    if args.input_dir:
        files = list(glob(os.path.join(args.input_dir, "*.tif")))
        input_paths.extend(files)

    if not input_paths:
        raise RuntimeError(
            (
                "No input files found. "
                "You should pass individual input_image paths, or use --input-dir."
            )
        )
    _logger.info("Num. input files: %d", len(input_paths))

    spatial_filter(input_paths=input_paths, output_path=args.output, mode=args.mode, size=args.size)


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()
